@@ -1,12 +1,12 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2012-2018 Snowflake Computing Inc. All right reserved.
+# Copyright (c) 2012-2019 Snowflake Computing Inc. All right reserved.
 #
 import binascii
 import decimal
 import time
-from datetime import datetime, timedelta, tzinfo, date
+from datetime import datetime, timedelta, date
 from logging import getLogger
 
 import pytz
@@ -17,8 +17,7 @@ from .errorcode import (
 from .errors import (ProgrammingError)
 from .sfbinaryformat import (binary_to_python,
                              binary_to_snowflake)
-from .sfdatetime import (sfdatetime_total_seconds_from_timedelta,
-                         sfdatetime_to_snowflake)
+from .sfdatetime import sfdatetime_total_seconds_from_timedelta
 
 try:
     import numpy
@@ -35,9 +34,6 @@ ZERO_EPOCH_DATE = date(1970, 1, 1)
 ZERO_EPOCH = datetime.utcfromtimestamp(0)
 ZERO_FILL = u'000000000'
 
-# Tzinfo class cache
-_TZINFO_CLASS_CACHE = {}
-
 logger = getLogger(__name__)
 
 PYTHON_TO_SNOWFLAKE_TYPE = {
@@ -50,7 +46,8 @@ PYTHON_TO_SNOWFLAKE_TYPE = {
     u'bytes': u'BINARY',
     u'bytearray': u'BINARY',
     u'bool': u'BOOLEAN',
-    u'nonetype': None,
+    u'bool_': u'BOOLEAN',
+    u'nonetype': u'ANY',
     u'datetime': u'TIMESTAMP_NTZ',
     u'sfdatetime': u'TIMESTAMP_NTZ',
     u'date': u'DATE',
@@ -102,15 +99,45 @@ def _convert_time_to_epoch_nanoseconds(tm):
            "{:06d}".format(tm.microsecond) + u'000'
 
 
+def _extract_timestamp(value, ctx):
+    """
+    Extracts timestamp from a raw data
+    """
+    scale = ctx['scale']
+    microseconds = float(
+        value[0:-scale + 6]) if scale > 6 else float(value)
+    fraction_of_nanoseconds = _adjust_fraction_of_nanoseconds(
+        value, ctx['max_fraction'], scale)
+
+    return microseconds, fraction_of_nanoseconds
+
+
+def _adjust_fraction_of_nanoseconds(value, max_fraction, scale):
+    if scale == 0:
+        return 0
+    if value[0] != '-':
+        return int((value[-scale:] + ZERO_FILL[:9 - scale]))
+
+    frac = int(value[-scale:])
+    if frac == 0:
+        return 0
+    else:
+        return int(TO_UNICODE(max_fraction - frac) + ZERO_FILL[:9 - scale])
+
+
+def _generate_tzinfo_from_tzoffset(tzoffset_minutes):
+    """
+    Generates tzinfo object from tzoffset.
+    """
+    return pytz.FixedOffset(tzoffset_minutes)
+
+
 class SnowflakeConverter(object):
     def __init__(self, **kwargs):
         self._parameters = {}
-        self._use_sfbinaryformat = kwargs.get('use_sfbinaryformat', False)
         self._use_numpy = kwargs.get('use_numpy', False) and numpy is not None
 
-        logger.debug('use_sfbinaryformat: %s, use_numpy: %s',
-                     self._use_sfbinaryformat,
-                     self._use_numpy)
+        logger.debug('use_numpy: %s', self._use_numpy)
 
     def set_parameters(self, parameters):
         self._parameters = {}
@@ -126,46 +153,6 @@ class SnowflakeConverter(object):
     def get_parameter(self, param):
         return self._parameters[param] if param in self._parameters else None
 
-    @staticmethod
-    def _adjust_fraction_of_nanoseconds(value, max_fraction, scale):
-        if scale == 0:
-            return 0
-        if value[0] != '-':
-            return int((value[-scale:] + ZERO_FILL)[:9])
-
-        return int((TO_UNICODE(max_fraction - int(value[-scale:])) +
-                    ZERO_FILL)[:9])
-
-    @staticmethod
-    def _generate_tzinfo_from_tzoffset(tzoffset_minutes):
-        """
-        Generates tzinfo object from tzoffset.
-        """
-        try:
-            return _TZINFO_CLASS_CACHE[tzoffset_minutes]
-        except KeyError:
-            pass
-        sign = u'P' if tzoffset_minutes >= 0 else u'N'
-        abs_tzoffset_minutes = abs(tzoffset_minutes)
-        hour, minute = divmod(abs_tzoffset_minutes, 60)
-        name = u'GMT{sign:s}{hour:02d}{minute:02d}'.format(
-            sign=sign,
-            hour=hour,
-            minute=minute)
-        tzinfo_class_type = type(
-            str(name),  # str() for both Python 2 and 3
-            (tzinfo,),
-            dict(
-                utcoffset=lambda self0, dt, is_dst=False: timedelta(
-                    minutes=tzoffset_minutes),
-                tzname=lambda self0, dt: name,
-                dst=lambda self0, dt: ZERO_TIMEDELTA
-            )
-        )
-        tzinfo_cls = tzinfo_class_type()
-        _TZINFO_CLASS_CACHE[tzoffset_minutes] = tzinfo_cls
-        return tzinfo_cls
-
     #
     # FROM Snowflake to Python Objects
     #
@@ -192,7 +179,7 @@ class SnowflakeConverter(object):
 
     def _FIXED_numpy_to_python(self, ctx):
         if ctx['scale']:
-            return decimal.Decimal
+            return numpy.float64
         else:
 
             def conv(value):
@@ -225,7 +212,7 @@ class SnowflakeConverter(object):
         def conv(value):
             try:
                 return datetime.utcfromtimestamp(int(value) * 86400).date()
-            except OSError as e:
+            except (OSError, ValueError) as e:
                 logger.debug("Failed to convert: %s", e)
                 ts = ZERO_EPOCH + timedelta(
                     seconds=int(value) * (24 * 60 * 60))
@@ -241,18 +228,6 @@ class SnowflakeConverter(object):
         """
         return lambda x: numpy.datetime64(int(x), 'D')
 
-    def _extract_timestamp(self, value, ctx):
-        """
-        Extracts timestamp from a raw data
-        """
-        scale = ctx['scale']
-        microseconds = float(
-            value[0:-scale + 6]) if scale > 6 else float(value)
-        fraction_of_nanoseconds = SnowflakeConverter._adjust_fraction_of_nanoseconds(
-            value, ctx['max_fraction'], scale)
-
-        return microseconds, fraction_of_nanoseconds
-
     def _TIMESTAMP_TZ_to_python(self, ctx):
         """
         TIMESTAMP TZ to datetime
@@ -264,43 +239,17 @@ class SnowflakeConverter(object):
 
         def conv0(encoded_value):
             value, tz = encoded_value.split()
-            tzinfo = SnowflakeConverter._generate_tzinfo_from_tzoffset(
-                int(tz) - 1440)
+            tzinfo = _generate_tzinfo_from_tzoffset(int(tz) - 1440)
             return datetime.fromtimestamp(float(value), tz=tzinfo)
 
         def conv(encoded_value):
             value, tz = encoded_value.split()
             microseconds = float(value[0:-scale + 6])
-            tzinfo = SnowflakeConverter._generate_tzinfo_from_tzoffset(
-                int(tz) - 1440)
+            tzinfo = _generate_tzinfo_from_tzoffset(int(tz) - 1440)
             return datetime.fromtimestamp(microseconds, tz=tzinfo)
 
         return conv if scale > 6 else conv0
 
-    def _TIMESTAMP_TZ_numpy_to_python(self, ctx):
-        """TIMESTAMP TZ to datetime
-
-        The timezone offset is piggybacked.
-        """
-        scale = ctx['scale']
-        zero_fill = ctx['zero_fill']
-
-        def conv0(encoded_value):
-            value, tz = encoded_value.split()
-            ts = 60 * (int(tz) - 1440) + int(value)
-            return numpy.datetime64(ts, 's')
-
-        def conv(encoded_value):
-            value, tz = encoded_value.split()
-            ts = (60 * (int(tz) - 1440) + int(value[0:-scale - 1])
-                  ) * 1000000000
-            if scale > 0:
-                ts += (-1 if value[0] == u'-' else 1) * int(
-                    value[-scale:] + zero_fill)
-            return numpy.datetime64(ts, 'ns')
-
-        return conv if scale > 0 else conv0
-
     def _get_session_tz(self):
         """ Get the session timezone or use the local computer's timezone. """
         try:
@@ -325,8 +274,7 @@ class SnowflakeConverter(object):
         This takes consideration of the session parameter TIMEZONE if
         available. If not, tzlocal is used
         """
-        microseconds, fraction_of_nanoseconds = \
-            self._extract_timestamp(value, ctx)
+        microseconds, fraction_of_nanoseconds = _extract_timestamp(value, ctx)
         tzinfo_value = self._get_session_tz()
 
         try:
@@ -338,7 +286,7 @@ class SnowflakeConverter(object):
                 "OverflowError in converting from epoch time to "
                 "timestamp_ltz: %s(ms). Falling back to use struct_time."
             )
-            return time.gmtime(microseconds), fraction_of_nanoseconds
+            return time.localtime(microseconds), fraction_of_nanoseconds
 
     def _TIMESTAMP_LTZ_to_python(self, ctx):
         tzinfo = self._get_session_tz()
@@ -352,29 +300,6 @@ class SnowflakeConverter(object):
 
         return conv if scale > 6 else conv0
 
-    def _TIMESTAMP_LTZ_numpy_to_python(self, ctx):
-        tzinfo = self._get_session_tz()
-        scale = ctx['scale']
-
-        def conv(value):
-            zero_fill = ctx['zero_fill']
-            seconds = int(value[0:-scale - 1]) if scale > 0 else int(value)
-            # construct datetime object to get utcoffset
-            dt = ZERO_EPOCH + timedelta(seconds=seconds)
-            offset = tzinfo.utcoffset(dt)
-            if offset.days < 0:
-                ts = (int(value[0:-scale - 1]) + (offset.seconds - 86400)
-                      ) * 1000000000
-            else:
-                ts = (int(value[0:-scale - 1]) + offset.seconds
-                      ) * 1000000000
-            if scale > 0:
-                ts += (-1 if value[0] == u'-' else 1) * int(
-                    value[-scale:] + zero_fill)
-            return numpy.datetime64(ts, 'ns')
-
-        return conv
-
     _TIMESTAMP_to_python = _TIMESTAMP_LTZ_to_python
 
     def _TIMESTAMP_NTZ_to_python(self, ctx):
@@ -569,6 +494,9 @@ class SnowflakeConverter(object):
     def _bool_to_snowflake(self, value):
         return value
 
+    def _bool__to_snowflake(self, value):
+        return bool(value)
+
     def _nonetype_to_snowflake(self, _):
         return None
 
@@ -579,7 +507,10 @@ class SnowflakeConverter(object):
         tzinfo_value = value.tzinfo
         if tzinfo_value:
             if pytz.utc != tzinfo_value:
-                td = tzinfo_value.utcoffset(value, is_dst=False)
+                try:
+                    td = tzinfo_value.utcoffset(value)
+                except pytz.exceptions.AmbiguousTimeError:
+                    td = tzinfo_value.utcoffset(value, is_dst=False)
             else:
                 td = ZERO_TIMEDELTA
             sign = u'+' if td >= ZERO_TIMEDELTA else u'-'
@@ -622,9 +553,6 @@ class SnowflakeConverter(object):
                 second=value.second
             )
 
-    def _sfdatetime_to_snowflake(self, value):
-        return sfdatetime_to_snowflake(value)
-
     def date_to_snowflake(self, value):
         """
         Converts Date object to Snowflake object
@@ -641,8 +569,7 @@ class SnowflakeConverter(object):
         return value.strftime(u'%H:%M:%S')
 
     def _struct_time_to_snowflake(self, value):
-        tzinfo_value = SnowflakeConverter._generate_tzinfo_from_tzoffset(
-            -time.timezone // 60)
+        tzinfo_value = _generate_tzinfo_from_tzoffset(time.timezone // 60)
         t = datetime.fromtimestamp(time.mktime(value))
         if pytz.utc != tzinfo_value:
             t += tzinfo_value.utcoffset(t)
@@ -691,7 +618,7 @@ class SnowflakeConverter(object):
     _float64_to_snowflake = __numpy_to_snowflake
 
     def _datetime64_to_snowflake(self, value):
-        return TO_UNICODE(value)
+        return TO_UNICODE(value) + u'+00:00'
 
     def _quoted_name_to_snowflake(self, value):
         return TO_UNICODE(value)
